@@ -2,6 +2,8 @@
 # Author: Dmitry Shachnev <mitya57@gmail.com>
 # License: GNU GPL 3 or higher; http://www.gnu.org/licenses/gpl.html
 
+from __future__ import print_function # pyflakes
+
 app_name = 'unity-mail'
 app_version = '1.2'
 
@@ -15,13 +17,16 @@ import sys
 import shutil
 import signal
 import time
+import subprocess
+import secretstorage
 
 from socket import error as socketerror
-from subprocess import Popen
-from gi.repository import GObject, GLib, Gtk, Notify, GnomeKeyring
+from gi.repository import GLib, Gtk, Notify
+from secretstorage.exceptions import SecretServiceNotAvailableException
 
-from UnityMail import *
-from UnityMail.config import *
+from UnityMail import gettext, Unity, Dbusmenu, MessagingMenu, \
+ with_unity, with_messagingmenu
+from UnityMail.config import ConfigFile, config_dir, open_url_or_command
 from UnityMail.dialog import PreferencesDialog
 
 _ = gettext.gettext
@@ -65,15 +70,6 @@ def get_sender_name(sender):
 	sname = email.utils.parseaddr(sender)[0]
 	return sname if sname else sender
 
-def get_value_from_attributes_list(attributes, name):
-	for attribute in attributes:
-		if attribute.name == name:
-			if attribute.type == GnomeKeyring.AttributeType.STRING:
-				return attribute.get_string()
-			else:
-				return attribute.get_uint32()
-
-get_item_id = lambda key: key.item_id
 fix_format = lambda string: string.replace('%(t0)s', '{t0}').replace('%(t1)s', '{t1}')
 
 class TimeoutException(Exception):
@@ -134,9 +130,9 @@ class UnityMail(object):
 			self.mmapp.register()
 			self.mmapp.connect('activate-source', self.on_mm_item_clicked)
 		
-		self.loop = GObject.MainLoop()
-		GObject.set_application_name('unity-mail')
-		GObject.timeout_add_seconds(self.interval, self.update)
+		self.loop = GLib.MainLoop()
+		GLib.set_application_name('unity-mail')
+		GLib.timeout_add_seconds(self.interval, self.update)
 		self.update()
 		self.start_loop()
 	
@@ -161,22 +157,26 @@ class UnityMail(object):
 	def on_mm_item_clicked(self, mmapp, sourceid):
 		for message in self.unread_messages:
 			if message.message_id == sourceid:
-				self.mark_message_as_read(message)
 				urlid = self.get_urlid(message.account_id)
 				if urlid:
 					open_url_or_command(urlid)
+				else:
+					self.mark_message_as_read(message)
 				if message.is_active and with_unity:
 					count = self.launcher.get_property('count')
 					self.set_launcher_count(count-1)
 	
 	def mark_message_as_read(self, message):
 		client = self.mail_client[message.account_id]
-		if message.folder:
-			client.select(message.folder)
-		else:
-			client.select()
-		client.store(message.account_position, '+FLAGS', '\Seen')
-		client.close()
+		try:
+			if message.folder:
+				client.select(message.folder)
+			else:
+				client.select()
+			client.store(message.account_position, '+FLAGS', '\Seen')
+			client.close()
+		except (imaplib.IMAP4.error, socketerror) as e:
+			print_error(e)
 	
 	def on_mark_all_read(self, menuitem, timestamp):
 		for message in self.unread_messages:
@@ -190,11 +190,13 @@ class UnityMail(object):
 		config = ConfigFile(config_dir+'/unity-mail.conf')
 		self.interval = config.get_int_with_default('Interval', 30)
 		self.enable_notifications = config.get_bool_with_default('EnableNotifications', True)
+		self.play_sound = config.get_bool_with_default('EnableSound', False)
 		self.display_icons = config.get_bool_with_default('NotificationsHaveIcons', False)
 		self.draw_attention = config.get_bool_with_default('DrawAttention', True)
 		self.hide_count = config.get_bool_with_default('HideMessagesCount', False)
 		
 		self.command = config.get_str_with_default('ExecOnReceive')
+		self.custom_sound = config.get_str_with_default('CustomSound')
 		self.account_display = config.get_str_with_default('AccountDisplay', None)
 		value = config.get_str_with_default('Blacklist')
 		self.black_list = value.split(', ') if value else []
@@ -236,34 +238,53 @@ class UnityMail(object):
 				self.open_dialog(set_dicts=True)
 	
 	def init_keyring(self):
-		search_attrs = GnomeKeyring.Attribute.list_new()
-		GnomeKeyring.Attribute.list_append_string(search_attrs, 'application', 'unity-mail')
-		result, self.mail_keys = GnomeKeyring.find_items_sync(GnomeKeyring.ItemType.NETWORK_PASSWORD, search_attrs)
+		bus = secretstorage.dbus_init()
+		self.collection = secretstorage.Collection(bus)
+		try:
+			locked = self.collection.is_locked()
+		except SecretServiceNotAvailableException:
+			# Sleep a second and try to read the collection by alias
+			time.sleep(1)
+			try:
+				self.collection = secretstorage.get_collection_by_alias(
+					bus, 'default')
+				locked = self.collection.is_locked()
+			except SecretServiceNotAvailableException as e:
+				sys.exit(e)
+		if locked:
+			self.collection.unlock()
+		if self.collection.is_locked():
+			sys.exit('unity-mail: Failed to unlock the collection; exiting')
+		search_attrs = {'application': 'unity-mail'}
+		self.mail_keys = list(self.collection.search_items(search_attrs))
 		self.dialogshown = False
-		if result == GnomeKeyring.Result.NO_MATCH:
+		if not self.mail_keys:
 			self.open_dialog()
 			self.dialogshown = True
-		if result == GnomeKeyring.Result.CANCELLED:
-			sys.exit('unity-mail: Failed to unlock the keyring; exiting')
-		for key in sorted(self.mail_keys, key=get_item_id):
-			attributes = GnomeKeyring.Attribute.list_to_glist(key.attributes)
-			self.host.append(get_value_from_attributes_list(attributes, 'server'))
-			self.port.append(get_value_from_attributes_list(attributes, 'port'))
-			self.login.append(get_value_from_attributes_list(attributes, 'username'))
-			self.passwd.append(key.secret)
+		for key in sorted(self.mail_keys, key=secretstorage.Item._item_id):
+			attributes = key.get_attributes()
+			self.host.append(attributes['server'])
+			self.port.append(int(attributes['port']))
+			self.login.append(attributes['username'])
+			self.passwd.append(key.get_secret().decode('utf-8'))
 			self.dicts.append({'Host': self.host[-1], 'Port': self.port[-1],
 			'Login': self.login[-1], 'Passwd': self.passwd[-1]})
 	
-	def create_keyring_item(self, ind):
-		attrs = GnomeKeyring.Attribute.list_new()
-		GnomeKeyring.Attribute.list_append_string(attrs, 'application', 'unity-mail')
-		GnomeKeyring.Attribute.list_append_string(attrs, 'service', 'imap')
-		GnomeKeyring.Attribute.list_append_string(attrs, 'server', self.host[ind])
-		GnomeKeyring.Attribute.list_append_uint32(attrs, 'port', self.port[ind])
-		GnomeKeyring.Attribute.list_append_string(attrs, 'username', self.login[ind])
-		name = 'unity-mail-{0}'.format(ind)
-		GnomeKeyring.item_create_sync(None, GnomeKeyring.ItemType.NETWORK_PASSWORD,
-		name, attrs, self.passwd[ind], True)
+	def create_keyring_item(self, ind, update=False):
+		attrs = {
+			'application': 'unity-mail',
+			'service': 'imap',
+			'server': self.host[ind],
+			'port': str(self.port[ind]),
+			'username': self.login[ind]
+		}
+		label = 'unity-mail: '+self.login[ind]+' at '+self.host[ind]
+		if update:
+			self.mail_keys[ind].set_attributes(attrs)
+			self.mail_keys[ind].set_secret(self.passwd[ind])
+			self.mail_keys[ind].set_label(label)
+		else:
+			self.collection.create_item(label, attrs, self.passwd[ind], True)
 	
 	def open_dialog(self, set_dicts=False):
 		self.accdlg = PreferencesDialog()
@@ -278,12 +299,13 @@ class UnityMail(object):
 			dlg.save_all_settings()
 			self.dicts = dlg.datadicts
 			self.load_data_from_dicts()
-			# Remove all old keys
-			for key in self.mail_keys:
-				GnomeKeyring.item_delete_sync(key.keyring, key.item_id)
-			# Create all new keys
-			for i in range(len(self.host)):
-				self.create_keyring_item(i)
+			for index in range(len(self.host), len(self.mail_keys)):
+				# Remove old keys
+				self.mail_keys[index].delete()
+			for index in range(len(self.host)):
+				# Create new keys or update existing
+				self.create_keyring_item(index,
+				update=(index < len(self.mail_keys)))
 		dlg.destroy()
 		if (not self.dicts) or (not self.host):
 			sys.exit('unity-mail: No accounts registered; exiting')
@@ -319,9 +341,9 @@ class UnityMail(object):
 		except:
 			return False
 		try:
-			res = self.mail_client[cn].login(str(self.login[cn]), str(self.passwd[cn]))
-		except imaplib.IMAP4.error as e:
-			print_error(e.args[0].decode())
+			self.mail_client[cn].login(self.login[cn], self.passwd[cn])
+		except (imaplib.IMAP4.error, UnicodeEncodeError) as e:
+			print_error(e)
 			sys.exit(
 			'unity-mail: Invalid account data provided for connection #{0}; exiting'.format(cn))
 		else:
@@ -394,7 +416,7 @@ class UnityMail(object):
 					self.notifications_queue.append((sender, subj, cn))
 				if self.command and not self.first_run:
 					try:
-						Popen((self.command, sender, ilabel))
+						subprocess.call((self.command, sender, ilabel))
 					except OSError as e:
 						# File doesn't exist or is not executable
 						print_error('unity-mail: Command cannot be executed:')
@@ -405,8 +427,9 @@ class UnityMail(object):
 	
 	def check_email(self, cn, email):
 		login, host = email.split('@')
-		return (self.login[cn] in (login, email)
-			and self.host[cn].endswith(host))
+		if self.login[cn] == email:
+			return True
+		return (self.login[cn] == login) and self.host[cn].endswith(host)
 	
 	def get_urlid(self, cn):
 		for urlid, url in self.on_click_urls:
@@ -493,7 +516,7 @@ class UnityMail(object):
 	
 	def set_launcher_count(self, count):
 		self.launcher.set_property('count', count)
-		if self.notinit == [True]*self.l:
+		if all(self.notinit):
 			# None of the connections is working
 			count_visible = False
 		else:
@@ -504,6 +527,14 @@ class UnityMail(object):
 		icon = 'unity-mail' if self.display_icons else None
 		basemessage = gettext.ngettext('You have %d unread mail', 'You have %d unread mails', number_of_mails)
 		basemessage = basemessage.replace('%d', '{0}')
+		if self.notifications_queue and self.play_sound:
+			try:
+				if self.custom_sound:
+					subprocess.call(('canberra-gtk-play', '-f', self.custom_sound))
+				else:
+					subprocess.call(('canberra-gtk-play', '-i', 'message-new-email'))
+			except OSError as e:
+				print(e)
 		if len(self.notifications_queue) > (1 if self.first_run else 2):
 			senders = set(get_sender_name(notification[0]) for notification in self.notifications_queue)
 			unknown_sender = ('' in senders)
@@ -518,7 +549,9 @@ class UnityMail(object):
 				message = _('from %s').replace('%s', '{0}').format(self.notifications_queue[0][0])
 			else:
 				message = None
-			Notify.Notification.new(basemessage.format(number_of_mails), message, icon).show()
+			notification = Notify.Notification.new(basemessage.format(number_of_mails), message, icon)
+			notification.set_hint_string('desktop-entry', 'unity-mail')
+			notification.show()
 		else:
 			for notification in self.notifications_queue:
 				if self.account_display is None:
@@ -532,4 +565,6 @@ class UnityMail(object):
 					message = _('New mail from %s').replace('%s', '{0}').format(notification[0])
 				else:
 					message = basemessage.format(1)
-				Notify.Notification.new(message, notification[1]+account_string, icon).show()
+				notification = Notify.Notification.new(message, notification[1]+account_string, icon)
+				notification.set_hint_string('desktop-entry', 'unity-mail')
+				notification.show()
